/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.examples.netDesignAlgorithm.fa;

import cern.colt.matrix.tdouble.DoubleMatrix2D;
import com.jom.OptimizationProblem;
import com.net2plan.interfaces.networkDesign.IAlgorithm;
import com.net2plan.interfaces.networkDesign.Net2PlanException;
import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.libraries.GraphUtils;
import com.net2plan.utils.CollectionUtils;
import com.net2plan.utils.Constants;
import com.net2plan.utils.Pair;
import com.net2plan.utils.Triple;
import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Given a network topology, the capacities in the links and the offered traffic, 
 * this algorithm obtains a link-disjoint or a SRG-disjoint (depending on user's choice) 
 * primary and backup path for the traffic of each demand, that minimizes the network 
 * congestion, with the constraint of non-bifurcated routing. For that it solves an 
 * ILP flow-link formulation (more complex in the SRG-disjoint case). In this algorithm, 
 * congestion is minimized by maximizing the idle link capacity in the bottleneck 
 * (the link with less idle capacity).
 *
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @version 1.2, September 2015
 */
public class FA_maxBottleneckIdleCapacity_xde_11pathProtection implements IAlgorithm
{
	@Override
	public String executeAlgorithm(NetPlan netPlan, Map<String, String> algorithmParameters, Map<String, String> net2planParameters)
	{
		/* Basic checks */
		final int N = netPlan.getNumberOfNodes();
		final int E = netPlan.getNumberOfLinks();
		final int D = netPlan.getNumberOfDemands();
		if (N == 0 || D == 0 || E == 0) throw new Net2PlanException("This algorithm requires a topology with links, and a demand set");
		
		final String type11 = algorithmParameters.get("type11");
		
		/* Initialize some variables */
		double[] u_e = netPlan.getLinkCapacityVector();
		double[] h_d = netPlan.getDemandOfferedTrafficVector();
		Set<Long> nodeIds = netPlan.getNodeIds();
		Map<Long, Pair<Long, Long>> linkMap = netPlan.getLinkMap();
		Map<Long, Pair<Long, Long>> demandMap = netPlan.getDemandMap();
		
		/* Create the optimization problem object (JOM library) */
		OptimizationProblem op = new OptimizationProblem();
		op.setInputParameter("u_e", u_e, "row");
		op.setInputParameter("h_d", h_d, "row");
		op.setInputParameter("A_ne", GraphUtils.getIncidenceMatrix(nodeIds, linkMap));
		op.setInputParameter("A_nd", GraphUtils.getIncidenceMatrix(nodeIds, demandMap));

		if (type11.equalsIgnoreCase("linkDisjoint"))
		{
			/* Add the decision variables to the problem */
			op.addDecisionVariable("x_de", true, new int[] { D, E }, 0, 1); /* 1 if the primary path of demand d, traverses link e */
			op.addDecisionVariable("xx_de", true, new int[] { D, E }, 0, 1);  /* 1 if the backup path of demand d, traverses link e */
			op.addDecisionVariable("u", false, new int[] { 1, 1 }, 0, Double.MAX_VALUE); /* idle capacity in the bottleneck link */
			
			/* Sets the objective function */
			op.setObjectiveFunction("maximize", "u");
			
			/* Add constraints */
			op.addConstraint("A_ne * (x_de') == A_nd"); /* the flow-conservation constraints for the primary path (NxD constraints) */
			op.addConstraint("A_ne * (xx_de') == A_nd"); /* the flow-conservation constraints for the backup path (NxD constraints) */
			op.addConstraint("x_de + xx_de <= 1"); /* the primary and backup path are link disjoint (DxE constraints) */
			op.addConstraint("h_d * (x_de + xx_de) <= u_e - u"); /* the capacity constraints summing primary and backup paths (E constraints) */
		}	
		else if (type11.equalsIgnoreCase("srgDisjoint"))
		{
			final int S = netPlan.getNumberOfSRGs();
			if (S == 0) throw new Net2PlanException("No SRG was defined");
			
			op.setInputParameter("A_se", computeAseMatrix(netPlan));
			op.setInputParameter("E", E);
			
			/* Add the decision variables to the problem */
			op.addDecisionVariable("x_de", true, new int[] { D, E }, 0, 1); /* 1 if the primary path of demand d, traverses link e */
			op.addDecisionVariable("xx_de", true, new int[] { D, E }, 0, 1);  /* 1 if the backup path of demand d, traverses link e */
			op.addDecisionVariable("u", false, new int[] { 1, 1 }, 0, Double.MAX_VALUE); /* idle capacity in the bottleneck link */
			op.addDecisionVariable("x_ds", true, new int[] { D, S }, 0, 1); /* 1 if demand d traverses SRG s in the primary  */
			op.addDecisionVariable("xx_ds", true, new int[] { D, S }, 0, 1); /* 1 if demand d traverses SRG s in the backup */
			
			/* Sets the objective function */
			op.setObjectiveFunction("maximize", "u");
			
			/* Add constraints */
			op.addConstraint("A_ne * (x_de') == A_nd"); /* the flow-conservation constraints for the primary path (NxD constraints) */
			op.addConstraint("A_ne * (xx_de') == A_nd"); /* the flow-conservation constraints for the backup path (NxD constraints) */
			op.addConstraint("h_d * (x_de + xx_de) <= u_e - u"); /* the capacity constraints summing primary and backup paths (E constraints) */
			op.addConstraint("E * x_ds >= x_de * A_se'"); /* 1 if demand s traverses SRG s in the primary (DxS constraints) */
			op.addConstraint("E * xx_ds >= xx_de * A_se'"); /* 1 if demand s traverses SRG s in the nackup (DxS constraints) */
			op.addConstraint("x_ds + xx_ds <= 1"); /* the primary and backup path are SRG disjoint (DxS constraints) */
		}
		else
		{
			throw new Net2PlanException("Wrong type of 1:1 protection");
		}
		
		/* Call the solver to solve the problem */
		String solverName = algorithmParameters.get("solverName");
		String solverLibraryName = algorithmParameters.get("solverLibraryName");
		op.solve(solverName, "solverLibraryName", solverLibraryName);

		/* If an optimal solution was not found, quit */
		if (!op.solutionIsOptimal()) throw new Net2PlanException("An optimal solution was not found");

		/* Retrieve the optimum solutions */
		DoubleMatrix2D x_de = op.getPrimalSolution("x_de").view2D();
		DoubleMatrix2D xx_de = op.getPrimalSolution("xx_de").view2D();

		/* Convert the x_de variables into a set of routes for each demand */
		List<Long> primary_demands = new ArrayList<Long>();
		List<List<Long>> primary_seqLinks = new ArrayList<List<Long>>();
		List<Double> primary_x_p = new ArrayList<Double>();
		GraphUtils.convert_xde2xp(linkMap, demandMap, x_de, primary_demands, primary_x_p, primary_seqLinks);

		List<Long> backup_demands = new ArrayList<Long>();
		List<List<Long>> backup_seqLinks = new ArrayList<List<Long>>();
		List<Double> backup_x_p = new ArrayList<Double>();
		GraphUtils.convert_xde2xp(linkMap, demandMap, xx_de, backup_demands, backup_x_p, backup_seqLinks);

		/* Remove all routes in current netPlan object */
		netPlan.removeAllRoutingInformation();
		netPlan.setRoutingType(Constants.RoutingType.SOURCE_ROUTING);

		/* Update netPlan object adding the calculated routes */
		if (primary_demands.size() != D) throw new Net2PlanException("Unexpected error");
		if (backup_demands.size() != D) throw new Net2PlanException("Unexpected error");
		
		for (long demandId : netPlan.getDemandIds())
		{
			/* for each demand, there is one primary and one backup path*/
			int primary_p = primary_demands.indexOf(demandId);
			if (primary_p == -1) throw new RuntimeException("Unexpected error");

			int backup_p = backup_demands.indexOf(demandId);
			if (backup_p == -1) throw new RuntimeException("Unexpected error");
			
			double h_d_thisDemand = netPlan.getDemandOfferedTraffic(demandId);
			long routeId = netPlan.addRoute(demandId, h_d_thisDemand, primary_seqLinks.get(primary_p), null); /* add the primary route (primary path) */
			long segmentId = netPlan.addProtectionSegment(backup_seqLinks.get(backup_p), h_d_thisDemand, null); /* add the protection segment (backup path) */
			netPlan.addProtectionSegmentToRouteBackupSegmentList(segmentId, routeId);
		}

		checkSolution(netPlan, type11);

		return "Ok!";
	}

	@Override
	public String getDescription()
	{
		return "Given a network topology, the capacities in the links and the offered traffic, this algorithm obtains a "
				+ "link-disjoint or SRG-disjoint (user's choice) primary and backup path for the traffic of each demand, that minimizes the network congestion, with the "
				+ "constraint of non-bifurcated routing, solving an ILP flow-link formulation. In this algorithm, congestion is minimized by maximizing the "
				+ "idle link capacity in the bottleneck (the link with less idle capacity).";
	}

	@Override
	public List<Triple<String, String, String>> getParameters()
	{
		List<Triple<String, String, String>> algorithmParameters = new LinkedList<Triple<String, String, String>>();
		algorithmParameters.add(Triple.of("solverName", "glpk", "The solver name to be used by JOM"));
		algorithmParameters.add(Triple.of("solverLibraryName", "", "The solver library full or relative path, to be used by JOM. Leave blank to use JOM default."));
		algorithmParameters.add(Triple.of("type11", "#select# linkDisjoint srgDisjoint", "Type of 1:1 protection: linkDisjoint or srgDisjoint. Formulations are somewhat different for both (SRG-disjoint is more complex)"));
		
		return algorithmParameters;
	}
	
	/**
	 * Creates the matrix associating SRGs and failed links (with 1).
	 * 
	 * @param netPlan Network design
	 * @return <i>A<sub>se</sub></i> matrix, which is equal to 1 if link <i>e</i> is associated to SRG <i>s</i> (or some of its end nodes), and zero otherwise
	 * @since 1.1
	 */
	public static double[][] computeAseMatrix(NetPlan netPlan)
	{
		final int S = netPlan.getNumberOfSRGs();
		final int E = netPlan.getNumberOfLinks();
		final double[][] A_se = new double[S][E];
		final Set<Long> linkIds = netPlan.getLinkIds();
		final Set<Long> srgIds = netPlan.getSRGIds();
		final Map<Long,Integer> linkId2LinearIndexMap = CollectionUtils.convertId2LinearIndexMap(linkIds);
		final Map<Long,Integer> srgId2LinearIndexMap = CollectionUtils.convertId2LinearIndexMap(srgIds);
		
		for (long srgId : netPlan.getSRGIds())
		{
			final int s = srgId2LinearIndexMap.get(srgId);
			for (long nodeId : netPlan.getSRGNodes(srgId))
			{
				for (long linkId : netPlan.getNodeAssociatedLinks(nodeId))
				{
					int e = linkId2LinearIndexMap.get(linkId);
					A_se[s][e] = 1;
				}
			}
			
			for (long linkId : netPlan.getSRGLinks(srgId))
			{
				int e = linkId2LinearIndexMap.get(linkId);
				A_se[s][e] = 1;
			}
		}		
		
		return A_se;
	}
	
	/**
	 * Checks if solution is valid.
	 * 
	 * @param netPlan Network design
	 * @param type11 Type of 1:1 protection
	 * @since 1.1
	 */
	private static void checkSolution(NetPlan netPlan, String type11)
	{
		if (!netPlan.getLinksOversubscribed().isEmpty()) throw new Net2PlanException("Bad - Some link is oversubscribed (constraint violated)");
		if (!netPlan.getDemandsBlocked().isEmpty()) throw new Net2PlanException("Bad - Some demand is blocked (constraint violated)");

		for (long routeId : netPlan.getRouteIds())
		{
			if (netPlan.getRouteBackupSegmentList(routeId).size () != 1) throw new RuntimeException("Bad");
			
			final long segmentId = netPlan.getRouteBackupSegmentList(routeId).get(0);
			if (netPlan.getRouteIngressNode(routeId) != netPlan.getProtectionSegmentOriginNode(segmentId)) throw new RuntimeException("Bad");
			if (netPlan.getRouteEgressNode(routeId) != netPlan.getProtectionSegmentDestinationNode(segmentId)) throw new RuntimeException("Bad");
			
			if (type11.equalsIgnoreCase("srgDisjoint"))
			{
				Set<Long> srgIds_primaryPath = netPlan.getRouteSRGs(routeId);
				Set<Long> srgIds_backupPath = netPlan.getProtectionSegmentSRGs(segmentId);
				if (!Collections.disjoint(srgIds_primaryPath, srgIds_backupPath))
					throw new RuntimeException("Bad");
			}
			else if (type11.equalsIgnoreCase("linkDisjoint"))
			{
				if (!Collections.disjoint(netPlan.getRouteSequenceOfLinks(routeId), netPlan.getProtectionSegmentSequenceOfLinks(segmentId)))
					throw new RuntimeException("Bad");
			}
			else
			{
				throw new RuntimeException ("Bad");
			}
		}
	}
}
